# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

from dataclasses import dataclass, field
from datetime import datetime
from typing import Optional

from air_sdk.v2.endpoints import mixins
from air_sdk.v2.air_model import AirModel, BaseEndpointApi
from air_sdk.v2.utils import validate_payload_types


@dataclass(eq=False)
class ApiToken(AirModel):
    id: str = field(repr=False)
    name: str
    created: datetime = field(repr=False)
    expiry: Optional[datetime]

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return ApiTokenEndpointApi


class ApiTokenEndpointApi(
    mixins.ListApiMixin[ApiToken],
    mixins.CreateApiMixin[ApiToken],
    mixins.DeleteApiMixin,
    BaseEndpointApi[ApiToken],
):
    API_PATH = 'api-tokens'
    model = ApiToken

    @validate_payload_types
    def create(self, name: str, expiry: Optional[datetime] = None) -> ApiToken:
        return super().create(name=name, expiry=expiry)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

import json
from dataclasses import dataclass, field
from datetime import datetime
from http import HTTPStatus
from io import TextIOBase, TextIOWrapper
from pathlib import Path
from typing import Callable, Dict, Literal, Optional, TypedDict, Union, cast

from air_sdk.util import raise_if_invalid_response
from air_sdk.v2.air_model import AirModel, BaseEndpointApi, DataDict, PrimaryKey
from air_sdk.v2.endpoints import mixins
from air_sdk.v2.endpoints.organizations import Organization
from air_sdk.v2.endpoints.workers import Worker
from air_sdk.v2.utils import join_urls, validate_payload_types

TopologyFormatContent = Union[Dict[str, object]]
TopologyFormatType = Literal['JSON']
SimulationImportPayloadContent = Union[str, TopologyFormatContent, Path, TextIOWrapper]


class SimulationImportResponse(TypedDict):
    id: str
    title: str
    organization: Optional[str]
    organization_name: Optional[str]


class TopologyFormat(TypedDict, total=False):
    title: str
    format: TopologyFormatType
    organization: Union[Organization, PrimaryKey]  # NotRequired
    content: TopologyFormatContent


@dataclass(eq=False)
class Simulation(AirModel):
    id: str
    title: str
    state: str
    sleep: bool = field(repr=False)
    owner: str = field(repr=False)
    cloned: bool = field(repr=False)
    expires: bool = field(repr=False)
    created: datetime = field(repr=False)
    modified: datetime = field(repr=False)
    sleep_at: datetime = field(repr=False)
    expires_at: datetime = field(repr=False)
    organization: Optional[Organization] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    preferred_worker: Optional[Worker] = field(metadata=AirModel.FIELD_FOREIGN_KEY, repr=False)
    documentation: Optional[str]
    write_ok: Optional[bool]
    metadata: Optional[str]

    @classmethod
    def get_model_api(cls):
        """
        Returns the respective `AirModelAPI` type for this model.
        """
        return SimulationEndpointApi

    @validate_payload_types
    def update(
        self,
        documentation: Optional[str] = None,
        expires: Optional[bool] = None,
        expires_at: Optional[datetime] = None,
        metadata: Optional[str] = None,
        preferred_worker: Optional[Union[Worker, PrimaryKey]] = None,
        sleep: Optional[bool] = None,
        sleep_at: Optional[datetime] = None,
        title: Optional[str] = None,
    ) -> None:
        data = {
            'documentation': documentation,
            'expires': expires,
            'expires_at': expires_at,
            'metadata': metadata,
            'preferred_worker': preferred_worker,
            'sleep': sleep,
            'sleep_at': sleep_at,
            'title': title,
        }
        data = {key: value for (key, value) in data.items() if value is not None}
        super().update(**data)

    @validate_payload_types
    def full_update(
        self,
        documentation: Optional[str],
        expires: bool,
        expires_at: datetime,
        metadata: Optional[str],
        preferred_worker: Optional[Union[Worker, PrimaryKey]],
        sleep: bool,
        sleep_at: datetime,
        title: str,
    ) -> None:
        super().update(
            documentation=documentation,
            expires=expires,
            expires_at=expires_at,
            metadata=metadata,
            preferred_worker=preferred_worker,
            sleep=sleep,
            sleep_at=sleep_at,
            title=title,
        )

    @validate_payload_types
    def export(
        self,
        format: TopologyFormatType,
        image_ids: bool = False,
    ) -> TopologyFormatContent:
        """
        Exports this simulation into a desired topology format.

        :Arguments:
        - `format`: desired output format (currently supported format is `JSON`)
        - `image_ids`: whether to return image IDs instead of names
        """
        return self.__api__.simulations.export(self, format, image_ids)


class SimulationEndpointApi(
    mixins.ListApiMixin[Simulation],
    mixins.CreateApiMixin[Simulation],
    mixins.GetApiMixin[Simulation],
    mixins.PatchApiMixin[Simulation],
    mixins.DeleteApiMixin,
    BaseEndpointApi[Simulation],
):
    API_PATH = 'simulations'
    IMPORT_PATH = 'import'
    EXPORT_PATH = 'export'
    model = Simulation

    @validate_payload_types
    def create(
        self,
        title: str,
        documentation: Optional[str] = None,
        expires: Optional[bool] = None,
        expires_at: Optional[datetime] = None,
        metadata: Optional[str] = None,
        organization: Optional[Union[Organization, PrimaryKey]] = None,
        owner: Optional[str] = None,
        preferred_worker: Optional[Union[Worker, PrimaryKey]] = None,
        sleep: Optional[bool] = None,
        sleep_at: Optional[datetime] = None,
    ) -> Simulation:
        payload: DataDict = {'title': title}
        for key, value in (
            ('documentation', documentation),
            ('expires', expires),
            ('expires_at', expires_at),
            ('metadata', metadata),
            ('organization', organization),
            ('owner', owner),
            ('preferred_worker', preferred_worker),
            ('sleep', sleep),
            ('sleep_at', sleep_at),
        ):
            if value is not None:  # Don't include optional None fields in payload.
                payload[key] = value
        return super().create(**payload)

    @staticmethod
    def _resolve_json_import_format_content(
        content: SimulationImportPayloadContent,
    ) -> Dict[str, object]:
        """
        Resolves given `content` into a JSON object.

        :Raises:
        - `JSONDecodeError` - `content` is not a valid JSON document
        - `ValueError` - resolved `content` is not a JSON object
        - `FileNotFoundError` - topology file does not exist at provided path
        """
        if isinstance(content, TextIOBase):
            resolved_content = json.load(content)
        elif isinstance(content, Path):
            with content.open('r') as content_file:
                resolved_content = json.load(content_file)
        elif isinstance(content, str):
            resolved_content = json.loads(content)
        else:
            resolved_content = content

        if not isinstance(resolved_content, dict):
            raise ValueError(
                f'JSON topology format payload content must be an object, got {type(resolved_content)}'
            )

        return resolved_content

    @validate_payload_types
    def create_from(
        self,
        title: str,
        format: TopologyFormatType,
        content: SimulationImportPayloadContent,
        organization: Optional[Union[Organization, PrimaryKey]] = None,
    ) -> Simulation:
        """
        Creates a simulation using a supported topology format. Currently supported format is `JSON`.

        :Raises:
        - `JSONDecodeError` - `content` is not a valid JSON document (when format is set to `JSON`)
        - `ValueError` - `content` value does not match `format`
        - `FileNotFoundError` - topology file does not exist at provided path
        """
        payload = cast(
            DataDict,
            TopologyFormat(
                title=title,
                format=format,
                content=IMPORT_FORMAT_CONTENT_HANDLERS[format](content),
            ),
        )
        if organization is not None:
            payload['organization'] = organization
        response = self.__api__.client.post(
            join_urls(self.url, self.IMPORT_PATH),
            data=mixins.serialize_payload(payload),
        )
        raise_if_invalid_response(response, status_code=HTTPStatus.CREATED)
        import_result = cast(SimulationImportResponse, response.json())

        return self.get(import_result['id'])

    @validate_payload_types
    def export(
        self,
        simulation: Union[Simulation, PrimaryKey],
        format: TopologyFormatType,
        image_ids: bool = False,
    ) -> TopologyFormatContent:
        """
        Exports an existing simulation into a desired topology format.

        :Arguments:
        - `format`: desired output format (currently supported format is `JSON`)
        - `image_ids`: whether to return image IDs instead of names
        """
        pk = str(simulation.__pk__ if isinstance(simulation, Simulation) else simulation)
        response = self.__api__.client.get(
            join_urls(self.url, pk, self.EXPORT_PATH),
            params={
                'topology_format': format,
                'image_ids': image_ids,
            },
        )
        raise_if_invalid_response(response, status_code=HTTPStatus.OK)

        return cast(TopologyFormat, response.json())['content']


IMPORT_FORMAT_CONTENT_HANDLERS: Dict[
    TopologyFormatType, Callable[[SimulationImportPayloadContent], TopologyFormatContent]
] = {
    'JSON': SimulationEndpointApi._resolve_json_import_format_content,
}



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
NVIDIA Air API module
"""

from datetime import date, datetime
from json import JSONDecodeError

import requests
from requests.compat import urlparse

from . import util, const
from .account import AccountApi
from .air_model import AirModel, LazyLoaded
from .capacity import CapacityApi
from .demo import DemoApi
from .exceptions import AirAuthorizationError, AirForbiddenError, AirUnexpectedResponse
from .fleet import FleetApi
from .image import ImageApi
from .interface import InterfaceApi
from .job import JobApi
from .link import LinkApi
from .logger import air_sdk_logger as logger
from .login import LoginApi
from .marketplace import MarketplaceApi
from .node import NodeApi
from .organization import OrganizationApi
from .permission import PermissionApi
from .resource_budget import ResourceBudgetApi
from .service import ServiceApi
from .simulation import SimulationApi
from .simulation_interface import SimulationInterfaceApi
from .simulation_node import SimulationNodeApi
from .ssh_key import SSHKeyApi
from .token import TokenApi
from .topology import TopologyApi
from .topology_file import TopologyFileApi
from .userconfig import UserConfigAPI
from .worker import WorkerApi


class AirSession(requests.Session):
    """Wrapper around requests.Session"""

    default_connect_timeout = const.DEFAULT_CONNECT_TIMEOUT
    default_read_timeout = const.DEFAULT_READ_TIMEOUT

    def rebuild_auth(self, prepared_request, response):
        """Allow credential sharing between nvidia.com and cumulusnetworks.com only"""
        if urlparse(prepared_request.url).hostname in const.ALLOWED_HOSTS:
            return
        super().rebuild_auth(prepared_request, response)

    def request(self, method, url, **kwargs):
        """Override request method to pass the timeout"""
        connect_timeout = kwargs.pop('default_connect_timeout', self.default_connect_timeout)
        read_timeout = kwargs.pop('default_read_timeout', self.default_read_timeout)
        kwargs.setdefault('timeout', (connect_timeout, read_timeout))
        return super().request(method, url, **kwargs)


class AirApi:
    """
    Main interface for an API client instance
    """

    def __init__(self, api_url=const.DEFAULT_API_URL, api_version='v1', **kwargs):
        """
        Create a new API client instance. The caller MUST provide either `username` and `password`
        or a `bearer_token`. The `password` argument may either be an API token or a service account
        password.

        Arguments:
            username (str, optional): Username
            password (str, optional): Password or API token
            bearer_token (str, optional): Pre-generated bearer token
            api_url (str, optional): Default = https://air.nvidia.com/api/
            api_version (str): Default = v1
        """
        self.client = AirSession()
        self.client.headers.update({'content-type': 'application/json'})

        self.api_url = _normalize_api_url(api_url) + _normalize_api_version(api_version)
        self._kwargs = kwargs
        self.token = None
        self.username = None
        self.authorize(**kwargs)

    @property
    def accounts(self):
        return AccountApi(self)

    @property
    def api_tokens(self):
        return TokenApi(self)

    @property
    def capacity(self):
        return CapacityApi(self)

    @property
    def demos(self):
        return DemoApi(self)

    @property
    def fleets(self):
        return FleetApi(self)

    @property
    def images(self):
        return ImageApi(self)

    @property
    def interfaces(self):
        return InterfaceApi(self)

    @property
    def jobs(self):
        return JobApi(self)

    @property
    def links(self):
        return LinkApi(self)

    @property
    def login(self):
        return LoginApi(self)

    @property
    def marketplace(self):
        return MarketplaceApi(self)

    @property
    @util.deprecated('AirApi.nodes')
    def node(self):
        return self.nodes

    @property
    def nodes(self):
        return NodeApi(self)

    @property
    def organizations(self):
        return OrganizationApi(self)

    @property
    @util.deprecated('AirApi.permissions')
    def permission(self):
        return self.permissions

    @property
    def permissions(self):
        return PermissionApi(self)

    @property
    def resource_budgets(self):
        return ResourceBudgetApi(self)

    @property
    @util.deprecated('AirApi.services')
    def service(self):
        return self.services

    @property
    def services(self):
        return ServiceApi(self)

    @property
    @util.deprecated('AirApi.simulations')
    def simulation(self):
        return self.simulations

    @property
    def simulations(self):
        return SimulationApi(self)

    @property
    @util.deprecated('AirApi.simulation_interfaces')
    def simulation_interface(self):
        return self.simulation_interfaces

    @property
    def simulation_interfaces(self):
        return SimulationInterfaceApi(self)

    @property
    @util.deprecated('AirApi.simulation_nodes')
    def simulation_node(self):
        return self.simulation_nodes

    @property
    def simulation_nodes(self):
        return SimulationNodeApi(self)

    @property
    def ssh_keys(self):
        return SSHKeyApi(self)

    @property
    @util.deprecated('AirApi.topologies')
    def topology(self):
        return self.topologies

    @property
    def topology_files(self):
        return TopologyFileApi(self)

    @property
    def topologies(self):
        return TopologyApi(self)

    @property
    @util.deprecated('AirApi.workers')
    def worker(self):
        return self.workers

    @property
    def workers(self):
        return WorkerApi(self)

    @property
    def user_configs(self):
        return UserConfigAPI(self)

    def authorize(self, **kwargs):
        """
        Authorizes the API client using either a pre-generated API token, a service account
        username/password, or a pre-generated bearer token.
        Callers MUST pass either a valid `bearer_token` or a `username` and `password`.
        The `password` argument may either be an API token or a service account
        password. After successfully authorizing, all subsequent API calls will include the
        authorization token provided by the AIR API. **Note:** This is called once automatically
        when an AirApi object is instantiated.

        Arguments:
            bearer_token (str, optional): Pre-generated bearer token
            username (str, optional): Username
            password (str, optional): Password or API token

        Raises:
            ValueError - Caller did not pass either a token or a username/password
        """
        token = None
        if kwargs.get('bearer_token'):
            token = kwargs['bearer_token']
        elif kwargs.get('username', None) and kwargs.get('password', None):
            token = self.get_token(kwargs['username'], kwargs['password'])
        else:
            raise ValueError('Must include either `bearer_token` or ' + '`username` and `password` arguments')
        self.token = token
        self.client.headers.update({'authorization': 'Bearer ' + token})
        login = self.login.list()
        self.username = getattr(login, 'username', None)

    def get_token(self, username, password):
        """
        Gets a new bearer token for a given username and password

        Arguments:
            username (str): Username
            password (str): Password

        Returns:
            str: Bearer token

        Raises:
            - [`AirAuthorizationError`](/docs/exceptions) - API did not return a token
            - `JSONDecodeError` - API's response is not a valid JSON object
        """
        route = '/login/'
        data = {'username': username, 'password': password}
        res = self.post(self.api_url + route, json=data, attempt_reauth=False)
        try:
            if res.json().get('token', None):
                return res.json()['token']
            logger.debug('AirApi.get_token :: Response JSON')
            logger.debug(res.json())
            raise AirAuthorizationError('API did not provide a token for ' + username)
        except JSONDecodeError:
            raise AirAuthorizationError('API did not return a valid JSON response')

    def _request(self, method, url, *args, **kwargs):
        attempt_reauth = kwargs.pop('attempt_reauth', True)
        if kwargs.get('json'):
            logger.debug(f'unserialized json: {kwargs["json"]}')
            if isinstance(kwargs['json'], list):
                kwargs['json'] = [_serialize_dict(obj) for obj in kwargs['json']]
            else:
                kwargs['json'] = _serialize_dict(kwargs['json'])
        if kwargs.get('params'):
            kwargs['params'] = _serialize_dict(kwargs['params'])
        logger.debug(f'request args: {args}')
        logger.debug(f'request kwargs: {kwargs}')
        res = self.client.request(method, url, allow_redirects=False, *args, **kwargs)
        if res.status_code == 301 and urlparse(res.headers.get('Location')).hostname in const.ALLOWED_HOSTS:
            res = self.client.request(method, res.headers['Location'], *args, **kwargs)
        if getattr(res, 'status_code') == 403:
            missing_creds_err_msg = '{"detail":"Authentication credentials were not provided."}'
            if attempt_reauth and self._kwargs.get('username') and self._kwargs.get('password'):
                logger.debug('Request failed with 403, attempting reauth')
                self.authorize(**self._kwargs)
                return self._request(method, url, *args, **{'attempt_reauth': False, **kwargs})
            if res.text != missing_creds_err_msg:
                raise AirForbiddenError(res.text)
            else:
                raise AirForbiddenError
        try:
            res.raise_for_status()
        except requests.exceptions.HTTPError as err:
            raise AirUnexpectedResponse(err.response.text, err.response.status_code)
        return res

    def get(self, url, *args, **kwargs):
        """Wrapper method for GET requests"""
        return self._request('GET', url, *args, **kwargs)

    def post(self, url, *args, **kwargs):
        """Wrapper method for POST requests"""
        return self._request('POST', url, *args, **kwargs)

    def put(self, url, *args, **kwargs):
        """Wrapper method for PUT requests"""
        return self._request('PUT', url, *args, **kwargs)

    def patch(self, url, *args, **kwargs):
        """Wrapper method for PATCH requests"""
        return self._request('PATCH', url, *args, **kwargs)

    def delete(self, url, *args, **kwargs):
        """Wrapper method for DELETE requests"""
        return self._request('DELETE', url, *args, **kwargs)


def _normalize_api_version(version):
    try:
        version = int(version)
        version = f'v{version}'
    except Exception:
        pass
    return version


def _normalize_api_url(url: str) -> str:
    if url[-1] != '/':
        url += '/'
    if not url.endswith('api/'):
        url += 'api/'
    return url


def _serialize_dict(raw_dict):
    clone = {}
    for key, value in raw_dict.items():
        if isinstance(value, (AirModel, LazyLoaded)):
            clone[key] = value.id
        elif isinstance(value, dict):
            clone[key] = _serialize_dict(value)
        elif isinstance(value, list):
            clone[key] = _serialize_list(value)
        elif isinstance(value, (datetime, date)):
            clone[key] = value.isoformat()
        elif not key.startswith('_'):
            clone[key] = value
    return clone


def _serialize_list(raw_list):
    clone = []
    for item in raw_list:
        if isinstance(item, (AirModel, LazyLoaded)):
            clone.append(item.id)
        elif isinstance(item, dict):
            clone.append(_serialize_dict(item))
        elif isinstance(item, list):
            clone.append(_serialize_list(item))
        elif not str(item).startswith('_'):
            clone.append(item)
    return clone



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
SDK-specific exceptions
"""

from typing import Optional


class AirError(Exception):
    def __init__(
        self,
        message: str = 'An error occurred within the air_sdk.v2.AirApi',
        status_code: Optional[int] = None,
    ):
        self.status_code = status_code
        super().__init__(message)


class AirModelAttributeError(AirError):
    def __init__(
        self,
        message: str = 'An error occurred while accessing an AirModel attribute.',
        status_code: Optional[int] = None,
    ):
        self.message = message
        super().__init__(message=self.message, status_code=status_code)



# SPDX-FileCopyrightText: Copyright (c) 2022-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
# SPDX-License-Identifier: MIT

"""
Image module
"""

from . import util
from .air_model import AirModel
from .logger import air_sdk_logger as logger


class Image(AirModel):
    """
    Manage an Image

    ### delete
    Delete the image. Once successful, the object should no longer be used and will raise
    [`AirDeletedObject`](/docs/exceptions) when referenced.

    Raises:
    [`AirUnexpectedResposne`](/docs/exceptions) - Delete failed

    ### json
    Returns a JSON string representation of the image

    ### refresh
    Syncs the image with all values returned by the API

    ### update
    Update the image with the provided data

    Arguments:
        kwargs (dict, optional): All optional keyword arguments are applied as key/value
                pairs in the request's JSON payload
    """

    def copy(self, organization):
        """
        Make a copy of the image in another organization

        Arguments:
            organization (str | `Organization`): `Organization` or ID

        Returns:
        [`Image`](/docs/image)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Copy failed

        Example:
        ```
        >>> image = air.images.get('33d8a377-ef0a-4a0d-ac2a-076e32678e18')
        >>> target_org = air.organizations.get('b0e47509-4099-4e24-b96f-d1278d431f46')
        >>> image.copy(target_org)
        <Image cumulus-vx-5.4.0 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self._api.url}{self.id}/copy/'
        res = self._api.client.post(url, json={'organization': organization})
        util.raise_if_invalid_response(res, status_code=201)
        return Image(self._api, **res.json())

    def upload(self, filename):
        """
        Upload an image file

        Arguments:
            filename (str): Absolute path to the local image

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Upload failed
        """
        url = f'{self._api.url}{self.id}/upload/'
        with open(filename, 'rb') as image_file:
            res = self._api.client.put(url, default_read_timeout=300, data=image_file)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    @util.required_kwargs(['contact'])
    def publish(self, contact: str):
        """
        Publish the image for public use

        Arguments:
            contact (str): The email address for the contact person associated with this image.

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Publish failed
        Example:
        ```
        >>> image.publish(contact='contact@nvidia.com')
        ```
        """
        url = f'{self._api.url}{self.id}/publish/'
        res = self._api.client.put(url, json={'contact': contact})
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    def unpublish(self):
        """
        Unpublish the image from public use

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - Unpublish failed

        Example:
        ```
        >>> image.unpublish()
        ```
        """
        url = f'{self._api.url}{self.id}/publish/'
        res = self._api.client.delete(url)
        util.raise_if_invalid_response(res, status_code=204, data_type=None)

    def __repr__(self):
        if self._deleted or not self.name:
            return super().__repr__()
        return f'<Image {self.name} {self.id}>'


class ImageApi:
    """High-level interface for the Image API"""

    def __init__(self, client):
        self.client = client
        self.url = self.client.api_url + '/image/'

    def get(self, image_id, **kwargs):
        """
        Get an existing image

        Arguments:
            image_id (str): Image ID
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        [`Image`](/docs/image)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.images.get('3dadd54d-583c-432e-9383-a2b0b1d7f551')
        <Image cumulus-vx-4.2.1 3dadd54d-583c-432e-9383-a2b0b1d7f551>
        ```
        """
        url = f'{self.url}{image_id}/'
        res = self.client.get(url, params=kwargs)
        util.raise_if_invalid_response(res)
        return Image(self, **res.json())

    def list(self, **kwargs):
        # pylint: disable=line-too-long
        """
        List existing images

        Arguments:
            kwargs (dict, optional): All other optional keyword arguments are applied as query
                parameters/filters

        Returns:
        list

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> air.images.list()
        [<Image cumulus-vx-4.2.1 c51b49b6-94a7-4c93-950c-e7fa4883591>, <Image generic/ubuntu18.04 3134711d-015e-49fb-a6ca-68248a8d4aff>]
        ```
        """  # pylint: enable=line-too-long
        res = self.client.get(f'{self.url}', params=kwargs)
        util.raise_if_invalid_response(res, data_type=list)
        return [Image(self, **image) for image in res.json()]

    @util.required_kwargs(
        ['name', 'organization', 'version', 'default_username', 'default_password', 'cpu_arch']
    )
    def create(self, **kwargs):
        """
        Create a new image

        Arguments:
            name (str): Image name
            organization (str | `Organization`): `Organization` or ID
            filename (str, optional): Absolute path to the local file which should be uploaded
            kwargs (dict, optional): All other optional keyword arguments are applied as key/value
                pairs in the request's JSON payload

        Returns:
        [`Image`](/docs/image)

        Raises:
        [`AirUnexpectedResposne`](/docs/exceptions) - API did not return a 200 OK
            or valid response JSON

        Example:
        ```
        >>> image = air.images.create(name='my_image', filename='/tmp/my_image.qcow2', agent_enabled=False)
        >>> image
        <Image my_image 01298e0c-4ef1-43ec-9675-93160eb29d9f>
        >>> image.upload_status
        'COMPLETE'
        >>> alt_img = air.images.create(name='my_alt_img', filename='/tmp/alt_img.qcow2', agent_enabled=False)
        >>> alt_img.upload_status
        'FAILED'
        ```
        """
        res = self.client.post(self.url, json=kwargs)
        util.raise_if_invalid_response(res, status_code=201)
        image = Image(self, **res.json())
        filename = kwargs.get('filename')
        if filename:
            try:
                image.upload(filename)
            except util.AirUnexpectedResponse as err:
                logger.error(err.message)
            image.refresh()
        return image
